import logging
import traceback
from datetime import datetime, timezone

from django.core.management import BaseCommand

from ixp_tracker.conf import IXP_TRACKER_DATA_LOOKUP_FACTORY
from ixp_tracker.data_lookup import AdditionalDataSources, DefaultAdditionalDataSources, load_lookup
from ixp_tracker.importers import import_data
from ixp_tracker.stats import generate_stats

logger = logging.getLogger("ixp_tracker")


class Command(BaseCommand):
    help = "Updates IXP data"

    def add_arguments(self, parser):
        parser.add_argument("--reset-asns", action="store_true", default=False, help="Do a full reset of ASNs rather than incremental update")
        parser.add_argument("--backfill", type=str, default=None, help="The month you would like to backfill data for")

    def handle(self, *args, **options):
        try:
            logger.debug("Importing IXP data")
            data_lookup: AdditionalDataSources = load_lookup(IXP_TRACKER_DATA_LOOKUP_FACTORY) or DefaultAdditionalDataSources()
            reset = options["reset_asns"]
            backfill_date = options["backfill"]
            processing_date = None
            if backfill_date is None:
                import_data(data_lookup, reset)
            else:
                processing_date = datetime.strptime(backfill_date, "%Y%m").replace(tzinfo=timezone.utc)
                if reset:
                    logger.warning("The --reset option has no effect when running a backfill")
                import_data(data_lookup, False, processing_date)

            logger.debug("Generating stats")
            generate_stats(data_lookup, processing_date)
            logger.info("Import finished")
        except Exception as e:
            logging.error("Failed to import data", extra={"error": str(e), "trace": traceback.format_exc()})



# Generated by Django 5.0.8 on 2025-08-27 07:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0012_statspercountry_routed_asn_count_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='ixp',
            name='manrs_participant',
            field=models.BooleanField(default=False),
        ),
    ]



from datetime import datetime, timezone

import pytest

from ixp_tracker import importers
from ixp_tracker.models import IXP
from tests.fixtures import IXPFactory, MockLookup, PeeringIXFactory

pytestmark = pytest.mark.django_db


def test_with_no_data_does_nothing():
    importers.process_ixp_data(datetime.now(timezone.utc), MockLookup())([])

    ixps = IXP.objects.all()
    assert len(ixps) == 0


def test_imports_a_new_ixp():
    importers.process_ixp_data(datetime.now(timezone.utc), MockLookup())([PeeringIXFactory()])

    ixps = IXP.objects.all()
    assert len(ixps) == 1
    ixp = ixps.first()
    assert not ixp.manrs_participant


def test_updates_an_existing_ixp():
    new_data = PeeringIXFactory()
    manrs_participants = [new_data["id"]]
    IXPFactory(peeringdb_id=new_data["id"], last_active=datetime(year=2024, month=4, day=1).replace(tzinfo=timezone.utc), manrs_participant=False)

    importers.process_ixp_data(datetime.now(timezone.utc), MockLookup(manrs_participants=manrs_participants))([new_data])

    ixps = IXP.objects.all()
    assert len(ixps) == 1
    ixp = IXP.objects.first()
    assert ixp.last_active.date() == datetime.now(timezone.utc).date()
    assert ixp.name == new_data["name"]
    assert ixp.manrs_participant


def test_does_not_import_an_ixp_from_a_non_iso_country():
    new_data = PeeringIXFactory()
    new_data["country"] = "XK"  # XK is Kosovo, but it's not an official ISO code
    importers.process_ixp_data(datetime.now(timezone.utc), MockLookup())([new_data])

    ixps = IXP.objects.all()
    assert len(ixps) == 0


def test_handles_errors_with_source_data():
    data_with_problems = PeeringIXFactory()
    data_with_problems["created"] = "abc"

    importers.process_ixp_data(datetime.now(timezone.utc), MockLookup())([data_with_problems])

    ixps = IXP.objects.all()
    assert len(ixps) == 0


def test_saves_manrs_participant():
    new_data = PeeringIXFactory()
    manrs_participants = [new_data["id"]]
    importers.process_ixp_data(datetime.now(timezone.utc), MockLookup(manrs_participants=manrs_participants))([new_data])

    ixp = IXP.objects.first()
    assert ixp.manrs_participant



# Generated by Django 5.2.5 on 2025-08-29 14:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0016_statspercountry_routed_asns_ixp_member_customers_rate_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='ixp',
            name='org_id',
            field=models.IntegerField(null=True),
        ),
    ]



from datetime import datetime, timedelta, timezone

import pytest

from ixp_tracker.models import StatsPerIXP
from ixp_tracker.stats import calculate_local_asns_members_rate, generate_stats
from tests.fixtures import ASNFactory, IXPFactory, IXPMembershipRecordFactory, MockLookup, StatsPerIXPFactory, \
    create_member_fixture

pytestmark = pytest.mark.django_db


def test_with_no_data_generates_no_stats():
    generate_stats(MockLookup())

    stats = StatsPerIXP.objects.all()
    assert len(stats) == 0


def test_generates_capacity_rs_peering_and_member_count():
    ixp = IXPFactory()
    create_member_fixture(ixp, membership_properties={"speed": 500, "is_rs_peer": True})
    create_member_fixture(ixp, membership_properties={"speed": 10000, "is_rs_peer": False})

    generate_stats(MockLookup())

    stats = StatsPerIXP.objects.all()
    assert len(stats) == 1
    ixp_stats = stats.first()
    assert ixp_stats.members == 2
    assert ixp_stats.capacity == 10.5
    assert ixp_stats.rs_peering_rate == 0.5


def test_generates_stats_for_first_of_month():
    IXPFactory()

    stats_date = datetime.now(timezone.utc).replace(day=10)
    generate_stats(MockLookup(), stats_date)

    stats = StatsPerIXP.objects.all()
    assert len(stats) == 1
    ixp_stats = stats.first()
    assert ixp_stats.stats_date == stats_date.replace(day=1).date()


def test_does_not_count_members_marked_as_left():
    ixp = IXPFactory()
    create_member_fixture(ixp, membership_properties={"speed": 500, "is_rs_peer": False})
    create_member_fixture(ixp, membership_properties={"speed": 10000, "is_rs_peer": True, "end_date": datetime(year=2024, month=4, day=1, tzinfo=timezone.utc)})

    generate_stats(MockLookup())

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.members == 1
    assert ixp_stats.capacity == 0.5
    assert ixp_stats.rs_peering_rate == 0


def test_does_not_count_member_twice_if_they_rejoin():
    ixp = IXPFactory()
    member = create_member_fixture(ixp, membership_properties={"end_date": datetime(year=2024, month=4, day=1, tzinfo=timezone.utc)})
    IXPMembershipRecordFactory(member=member)

    generate_stats(MockLookup())

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.members == 1


def test_does_not_count_members_not_yet_created():
    stats_date = datetime(year=2024, month=2, day=1, tzinfo=timezone.utc)
    ixp = IXPFactory(created=stats_date)
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2024, month=1, day=1)})
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2024, month=4, day=1)})

    generate_stats(MockLookup(), stats_date)

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.members == 1


def test_does_not_count_ixps_not_yet_created():
    stats_date = datetime(year=2024, month=2, day=1, tzinfo=timezone.utc)
    ixp = IXPFactory(created=(stats_date + timedelta(days=60)))
    create_member_fixture(ixp, quantity=2)

    generate_stats(MockLookup(), stats_date)

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats is None


def test_saves_local_asns_members_rate():
    ixp_one = IXPFactory()
    local_member_one = create_member_fixture(ixp_one)
    create_member_fixture(ixp_one, quantity=2)

    local_asns = [local_member_one.asn.number, ASNFactory().number, ASNFactory().number, ASNFactory().number]
    generate_stats(MockLookup(asns=local_asns))

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.local_asns_members_rate == 0.25


def test_saves_local_routed_asns_members_rate():
    ixp_one = IXPFactory()
    local_member_one = create_member_fixture(ixp_one)
    create_member_fixture(ixp_one, quantity=2)

    local_asns = [local_member_one.asn.number, ASNFactory().number, ASNFactory().number, ASNFactory().number]
    generate_stats(MockLookup(routed_asns=local_asns))

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.local_routed_asns_members_rate == 0.25


def test_calculate_local_asns_members_rate_returns_zero_if_no_asns_in_country():
    rate = calculate_local_asns_members_rate([12345], [])

    assert rate == 0


def test_calculate_local_asns_members_rate():
    rate = calculate_local_asns_members_rate([12345], [12345, 446, 789, 5050, 54321])

    assert rate == 0.2


def test_calculate_local_asns_members_rate_ignores_members_not_in_country_list():
    rate = calculate_local_asns_members_rate([12345, 789], [12345, 446, 5050, 54321])

    assert rate == 0.25


def test_counts_net_joins_and_net_leaves_since_12_months():
    stats_date = datetime(year=2024, month=2, day=1, tzinfo=timezone.utc)
    ixp = IXPFactory(created=stats_date)
    # One member joined more than 12 months ago and is still a member (i.e. not counted in either)
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2023, month=1, day=1)})
    # Two members joined within the last 12 months
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2024, month=1, day=1)})
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2023, month=12, day=1)})
    # One member joined more than 12 months ago but has since left
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2022, month=11, day=1), "end_date": datetime(year=2023, month=6, day=17)})
    # One member left and rejoined within the 12 months (so should not be counted)
    member = create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2022, month=11, day=1), "end_date": datetime(year=2023, month=6, day=17)})
    IXPMembershipRecordFactory(member=member, start_date=datetime(year=2023, month=11, day=1))

    generate_stats(MockLookup(), stats_date)

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.members_joined_last_12_months == 2
    assert ixp_stats.members_left_last_12_months == 1


def test_adds_member_growth_stats():
    stats_date = datetime(year=2025, month=3, day=1, tzinfo=timezone.utc)
    ixp = IXPFactory(created=stats_date)
    # Has 5 current members
    create_member_fixture(ixp, membership_properties={"start_date": datetime(year=2023, month=1, day=1)}, quantity=5)
    last_month = (stats_date - timedelta(days=1)).replace(day=1).date()
    # Note the fixture would have had 5 members last month but we hardcode the stats to show only 4
    StatsPerIXPFactory(ixp=ixp, members=4, stats_date=last_month)

    generate_stats(MockLookup(), stats_date)

    new_ixp_stats = StatsPerIXP.objects.filter(stats_date=stats_date.date()).first()
    assert new_ixp_stats.monthly_members_change == 1
    assert new_ixp_stats.monthly_members_change_percent == 0.25


def test_saves_local_routed_asns_members_and_customers_rate():
    ixp_one = IXPFactory()
    local_member_one = create_member_fixture(ixp_one)
    create_member_fixture(ixp_one, quantity=2)
    customer_asn = ASNFactory().number

    local_asns = [local_member_one.asn.number, customer_asn, ASNFactory().number, ASNFactory().number]
    generate_stats(MockLookup(routed_asns=local_asns, customer_asns=[customer_asn]))

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.local_routed_asns_members_customers_rate == 0.5



# Generated by Django 5.2.5 on 2025-08-27 12:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0014_statsperixp_members_joined_last_12_months_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='statsperixp',
            name='monthly_members_change',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='statsperixp',
            name='monthly_members_change_percent',
            field=models.FloatField(default=0),
            preserve_default=False,
        ),
    ]
